# -*- coding: utf-8 -*-
# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import datetime
from decimal import Decimal

from proteus import Model

__all__ = ['get_currency']

_names = {
    'USD': 'U.S. Dollar',
    'EUR': 'Euro',
    }
_symbols = {
    'USD': '$',
    'EUR': '€',
    }
_rates = {
    'USD': Decimal('1.0'),
    'EUR': Decimal('2.0'),
    }


def get_currency(code='USD', config=None):
    "Get currency with code"
    Currency = Model.get('currency.currency', config=config)
    CurrencyRate = Model.get('currency.currency.rate', config=config)

    currencies = Currency.find([('code', '=', code)])
    if not currencies:
        currency = Currency(name=_names.get(code, code),
            symbol=_symbols.get(code), code=code,
            rounding=Decimal('0.01'))
        currency.save()
        rate = _rates.get(code)
        if rate is not None:
            CurrencyRate(date=datetime.date.min, rate=rate,
                currency=currency).save()
    else:
        currency, = currencies
    return currency



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

from .test_module import add_currency_rate, create_currency

__all__ = ['create_currency', 'add_currency_rate']



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

import datetime
import unittest
from decimal import ROUND_HALF_DOWN, Decimal

from trytond import backend
from trytond.modules.currency.ecb import (
    RatesNotAvailableError, UnsupportedCurrencyError, get_rates)
from trytond.pool import Pool
from trytond.tests.test_tryton import ModuleTestCase, with_transaction
from trytond.transaction import Transaction


def create_currency(name):
    pool = Pool()
    Currency = pool.get('currency.currency')
    return Currency.create([{
                'name': name,
                'symbol': name,
                'code': name,
                }])[0]


def add_currency_rate(currency, rate, date=datetime.date.min):
    pool = Pool()
    Rate = pool.get('currency.currency.rate')
    return Rate.create([{
                'currency': currency.id,
                'rate': rate,
                'date': date,
                }])[0]


class CurrencyTestCase(ModuleTestCase):
    'Test Currency module'
    module = 'currency'

    def get_currency(self, code):
        return self.currency.search([
            ('code', '=', code),
            ], limit=1)[0]

    @with_transaction()
    def test_currencies(self):
        'Create currencies'
        cu1 = create_currency('cu1')
        cu2 = create_currency('cu2')
        self.assertTrue(cu1)
        self.assertTrue(cu2)

    @with_transaction()
    def test_rate(self):
        'Create rates'
        cu1 = create_currency('cu1')
        cu2 = create_currency('cu2')

        rate1 = add_currency_rate(cu1, Decimal("1.3"))
        rate2 = add_currency_rate(cu2, Decimal("1"))
        self.assertTrue(rate1)
        self.assertTrue(rate2)

        self.assertEqual(cu1.rate, Decimal("1.3"))

    @with_transaction()
    def test_rate_unicity(self):
        'Rate unicity'
        pool = Pool()
        Rate = pool.get('currency.currency.rate')
        Date = pool.get('ir.date')
        today = Date.today()

        cu = create_currency('cu')

        Rate.create([{
                    'rate': Decimal("1.3"),
                    'currency': cu.id,
                    'date': today,
                    }])

        self.assertRaises(Exception, Rate.create, {
                'rate': Decimal("1.3"),
                'currency': cu.id,
                'date': today,
                })

    @with_transaction()
    def test_round(self):
        "Test simple round"
        cu = create_currency('cu')
        cu.rounding = Decimal('0.001')
        cu.digits = 3
        cu.save()

        rounded = cu.round(Decimal('1.2345678'))

        self.assertEqual(rounded, Decimal('1.235'))

    @with_transaction()
    def test_round_non_unity(self):
        "Test round with non unity"
        cu = create_currency('cu')
        cu.rounding = Decimal('0.02')
        cu.digits = 2
        cu.save()

        rounded = cu.round(Decimal('1.2345'))

        self.assertEqual(rounded, Decimal('1.24'))

    @with_transaction()
    def test_round_big_number(self):
        "Test rounding big number"
        cu = create_currency('cu')

        rounded = cu.round(Decimal('1E50'))

        self.assertEqual(rounded, Decimal('1E50'))

    @with_transaction()
    def test_round_negative(self):
        "Test rounding with negative rounding"
        cu = create_currency('cu')
        cu.rounding = -Decimal('0.1')
        cu.digits = 1
        cu.save()

        rounded = cu.round(Decimal('1.23'))

        self.assertEqual(rounded, Decimal('1.2'))

    @with_transaction()
    def test_round_zero(self):
        "Test rounding with 0 as rounding"
        cu = create_currency('cu')
        cu.rounding = Decimal('0')
        cu.save()

        rounded = cu.round(Decimal('1.2345'))

        self.assertEqual(rounded, Decimal('1.2345'))

    @with_transaction()
    def test_round_opposite(self):
        "Test the opposite rounding"
        cu = create_currency('cu')
        cu.save()

        rounded = cu.round(Decimal('1.235'))
        self.assertEqual(rounded, Decimal('1.24'))
        opposite_rounded = cu.round(Decimal('1.235'), opposite=True)
        self.assertEqual(opposite_rounded, Decimal('1.24'))

    @with_transaction()
    def test_round_opposite_HALF_DOWN(self):
        "Test the oposite rounding of ROUND_HALF_DOWN"
        cu = create_currency('cu')
        cu.save()

        rounded = cu.round(Decimal('1.235'), rounding=ROUND_HALF_DOWN)
        self.assertEqual(rounded, Decimal('1.23'))
        opposite_rounded = cu.round(
            Decimal('1.235'), rounding=ROUND_HALF_DOWN, opposite=True)
        self.assertEqual(opposite_rounded, Decimal('1.24'))

    @with_transaction()
    def test_is_zero(self):
        "Test is zero"
        cu = create_currency('cu')
        cu.rounding = Decimal('0.001')
        cu.digits = 3
        cu.save()

        for value, result in [
                (Decimal('0'), True),
                (Decimal('0.0002'), True),
                (Decimal('0.0009'), False),
                (Decimal('0.002'), False),
                ]:
            with self.subTest(value=value):
                self.assertEqual(cu.is_zero(value), result)
            with self.subTest(value=-value):
                self.assertEqual(cu.is_zero(-value), result)

    @with_transaction()
    def test_is_zero_negative(self):
        "Test is zero with negative rounding"
        cu = create_currency('cu')
        cu.rounding = Decimal('-0.001')
        cu.digits = 3
        cu.save()

        for value, result in [
                (Decimal('0'), True),
                (Decimal('0.0002'), True),
                (Decimal('0.0009'), False),
                (Decimal('0.002'), False),
                ]:
            with self.subTest(value=value):
                self.assertEqual(cu.is_zero(value), result)
            with self.subTest(value=-value):
                self.assertEqual(cu.is_zero(-value), result)

    @with_transaction()
    def test_is_zero_zero(self):
        "Test is zero with 0 as rounding"
        cu = create_currency('cu')
        cu.rounding = Decimal('0')
        cu.save()

        for value, result in [
                (Decimal('0'), True),
                (Decimal('0.0002'), False),
                (Decimal('0.0009'), False),
                (Decimal('0.002'), False),
                ]:
            with self.subTest(value=value):
                self.assertEqual(cu.is_zero(value), result)
            with self.subTest(value=-value):
                self.assertEqual(cu.is_zero(-value), result)

    @with_transaction()
    def test_compute_simple(self):
        'Simple conversion'
        pool = Pool()
        Currency = pool.get('currency.currency')
        cu1 = create_currency('cu1')
        cu2 = create_currency('cu2')
        add_currency_rate(cu1, Decimal("1.3"))
        add_currency_rate(cu2, Decimal("1"))

        amount = Decimal("10")
        expected = Decimal("13")
        converted_amount = Currency.compute(
            cu2, amount, cu1, True)
        self.assertEqual(converted_amount, expected)

    @with_transaction()
    def test_compute_nonfinite(self):
        'Conversion with rounding on non-finite decimal representation'
        pool = Pool()
        Currency = pool.get('currency.currency')
        cu1 = create_currency('cu1')
        cu2 = create_currency('cu2')
        add_currency_rate(cu1, Decimal("1.3"))
        add_currency_rate(cu2, Decimal("1"))

        amount = Decimal("10")
        expected = Decimal("7.69")
        converted_amount = Currency.compute(
            cu1, amount, cu2, True)
        self.assertEqual(converted_amount, expected)

    @with_transaction()
    def test_compute_nonfinite_worounding(self):
        'Same without rounding'
        pool = Pool()
        Currency = pool.get('currency.currency')
        cu1 = create_currency('cu1')
        cu2 = create_currency('cu2')
        add_currency_rate(cu1, Decimal("1.3"))
        add_currency_rate(cu2, Decimal("1"))

        amount = Decimal("10")
        expected = Decimal('7.692307692307692307692307692')
        converted_amount = Currency.compute(
            cu1, amount, cu2, False)
        self.assertEqual(converted_amount, expected)

    @with_transaction()
    def test_compute_same(self):
        'Conversion to the same currency'
        pool = Pool()
        Currency = pool.get('currency.currency')
        cu1 = create_currency('cu1')
        add_currency_rate(cu1, Decimal("1.3"))

        amount = Decimal("10")
        converted_amount = Currency.compute(
            cu1, amount, cu1, True)
        self.assertEqual(converted_amount, amount)

    @with_transaction()
    def test_compute_zeroamount(self):
        'Conversion with zero amount'
        pool = Pool()
        Currency = pool.get('currency.currency')
        cu1 = create_currency('cu1')
        cu2 = create_currency('cu2')
        add_currency_rate(cu1, Decimal("1.3"))
        add_currency_rate(cu2, Decimal("1"))

        expected = Decimal("0")
        converted_amount = Currency.compute(
            cu1, Decimal("0"), cu2, True)
        self.assertEqual(converted_amount, expected)

    @with_transaction()
    def test_compute_zerorate(self):
        'Conversion with zero rate'
        pool = Pool()
        Currency = pool.get('currency.currency')
        cu1 = create_currency('cu1')
        cu2 = create_currency('cu2')

        add_currency_rate(cu1, Decimal('0'))
        add_currency_rate(cu2, Decimal('1'))

        amount = Decimal("10")
        self.assertRaises(Exception, Currency.compute,
            cu1, amount, cu2, True)
        self.assertRaises(Exception, Currency.compute,
            cu2, amount, cu1, True)

    @with_transaction()
    def test_compute_missingrate(self):
        'Conversion with missing rate'
        pool = Pool()
        Currency = pool.get('currency.currency')
        cu1 = create_currency('cu1')
        cu3 = create_currency('cu3')
        add_currency_rate(cu1, Decimal("1.3"))

        amount = Decimal("10")
        self.assertRaises(Exception, Currency.compute,
            cu3, amount, cu1, True)
        self.assertRaises(Exception, Currency.compute,
            cu1, amount, cu3, True)

    @with_transaction()
    def test_compute_bothmissingrate(self):
        'Conversion with both missing rate'
        pool = Pool()
        Currency = pool.get('currency.currency')
        cu3 = create_currency('cu3')
        cu4 = create_currency('cu4')

        amount = Decimal("10")
        self.assertRaises(Exception, Currency.compute,
            cu3, amount, cu4, True)

    @with_transaction()
    def test_delete_cascade(self):
        'Test deletion of currency deletes rates'
        pool = Pool()
        Currency = pool.get('currency.currency')
        Rate = pool.get('currency.currency.rate')
        currencies = [create_currency('cu%s' % i) for i in range(3)]
        [add_currency_rate(c, Decimal('1')) for c in currencies]
        Currency.delete(currencies)

        rates = Rate.search([(
                    'currency', 'in', list(map(int, currencies)),
                    )], 0, None, None)
        self.assertFalse(rates)

    @with_transaction()
    def test_currency_rate_sql(self):
        "Test currency rate SQL"
        pool = Pool()
        Currency = pool.get('currency.currency')
        Rate = pool.get('currency.currency.rate')
        transaction = Transaction()
        cursor = transaction.connection.cursor()
        date = datetime.date

        cu1 = create_currency('cu1')
        for date_, rate in [
                (date(2017, 1, 1), Decimal(1)),
                (date(2017, 2, 1), Decimal(2)),
                (date(2017, 3, 1), Decimal(3))]:
            add_currency_rate(cu1, rate, date_)
        cu2 = create_currency('cu2')
        for date_, rate in [
                (date(2017, 2, 1), Decimal(2)),
                (date(2017, 4, 1), Decimal(4))]:
            add_currency_rate(cu2, rate, date_)

        query = Currency.currency_rate_sql()
        if backend.name == 'sqlite':
            query.columns[-1].output_name += (
                ' [%s]' % Rate.date.sql_type().base)
        cursor.execute(*query)
        data = set(cursor)
        result = {
            (cu1.id, Decimal(1), date(2017, 1, 1), date(2017, 2, 1)),
            (cu1.id, Decimal(2), date(2017, 2, 1), date(2017, 3, 1)),
            (cu1.id, Decimal(3), date(2017, 3, 1), None),
            (cu2.id, Decimal(2), date(2017, 2, 1), date(2017, 4, 1)),
            (cu2.id, Decimal(4), date(2017, 4, 1), None),
            }

        self.assertSetEqual(data, result)


class ECBtestCase(unittest.TestCase):

    def test_rate_EUR_week_ago(self):
        "Fetch EUR rates a week ago"
        week_ago = datetime.date.today() - datetime.timedelta(days=7)

        rates = get_rates('EUR', week_ago)

        self.assertNotIn('EUR', rates)
        self.assertIn('USD', rates)

    def test_rate_USD_week_ago(self):
        "Fetch USD rates a week ago"
        week_ago = datetime.date.today() - datetime.timedelta(days=7)

        rates = get_rates('USD', week_ago)

        self.assertIn('EUR', rates)
        self.assertNotIn('USD', rates)

    def test_rate_EUR_on_weekend(self):
        "Fetch EUR rates on weekend"

        rates_fr = get_rates('EUR', datetime.date(2022, 9, 30))
        rates_sa = get_rates('EUR', datetime.date(2022, 10, 2))
        rates_su = get_rates('EUR', datetime.date(2022, 10, 2))

        self.assertEqual(rates_sa, rates_fr)
        self.assertEqual(rates_su, rates_fr)

    def test_rate_USD_on_weekend(self):
        "Fetch USD rates on weekend"

        rates_fr = get_rates('USD', datetime.date(2022, 9, 30))
        rates_sa = get_rates('USD', datetime.date(2022, 10, 2))
        rates_su = get_rates('USD', datetime.date(2022, 10, 2))

        self.assertEqual(rates_sa, rates_fr)
        self.assertEqual(rates_su, rates_fr)

    def test_rate_EUR_start_date(self):
        "Fetch EUR rates at start date"

        rates = get_rates('EUR', datetime.date(1999, 1, 4))

        self.assertEqual(rates['USD'], Decimal('1.1789'))

    def test_rate_USD_start_date(self):
        "Fetch USD rates at start date"

        rates = get_rates('USD', datetime.date(1999, 1, 4))

        self.assertEqual(rates['EUR'], Decimal('0.8482'))

    def test_rate_in_future(self):
        "Fetch rate in future raise an error"
        future = datetime.date.today() + datetime.timedelta(days=2)

        with self.assertRaises(RatesNotAvailableError):
            get_rates('USD', future)

    def test_rate_unsupported_currency(self):
        "Fetch rate for unsupported currency"
        with self.assertRaises(UnsupportedCurrencyError):
            get_rates('XXX', datetime.date(2022, 10, 3))


del ModuleTestCase



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from trytond.config import config
from trytond.model import fields
from trytond.pool import PoolMeta

rate_decimal = config.getint('currency', 'rate_decimal', default=6)


class Configuration(metaclass=PoolMeta):
    __name__ = 'ir.configuration'
    currency_rate_decimal = fields.Integer("Currency Rate Decimal")

    @classmethod
    def default_currency_rate_decimal(cls):
        return rate_decimal

    def check(self):
        super().check()
        if self.currency_rate_decimal != rate_decimal:
            raise ValueError(
                "The rate_decimal %s in the [currency] configuration section "
                "is different from the value %s in 'ir.configuration'." % (
                    rate_decimal, self.currency_rate_decimal))


class Cron(metaclass=PoolMeta):
    __name__ = 'ir.cron'

    @classmethod
    def __setup__(cls):
        super().__setup__()
        cls.method.selection.append(
            ('currency.cron|update', "Update Currency Rates"))



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

from trytond.pool import Pool


def register():
    # Prevent to import backend when importing scripts
    from . import currency, ir
    Pool.register(
        currency.Currency,
        currency.CurrencyRate,
        currency.Cron,
        currency.Cron_Currency,
        ir.Cron,
        module='currency', type_='model')



# This file is part of Tryton.  The COPYRIGHT file at the toplevel of this
# repository contains the full copyright notices and license terms.
from trytond.model import fields

__all__ = ['Monetary']


class Monetary(fields.Numeric):
    """
    Define a numeric field with currency (``decimal``).
    """
    def __init__(self, string='', currency=None, digits=None, help='',
            required=False, readonly=False, domain=None, states=None,
            on_change=None, on_change_with=None, depends=None, context=None,
            loading='eager'):
        '''
        :param currency: the name of the Many2One field which stores
            the currency
        '''
        if currency:
            if depends is None:
                depends = set()
            else:
                depends = set(depends)
            depends.add(currency)
        super().__init__(string=string, digits=digits, help=help,
            required=required, readonly=readonly, domain=domain, states=states,
            on_change=on_change, on_change_with=on_change_with,
            depends=depends, context=context, loading=loading)
        self.currency = currency

    def definition(self, model, language):
        definition = super().definition(model, language)
        definition['symbol'] = self.currency
        definition['monetary'] = True
        return definition



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import datetime
import decimal
import logging
from decimal import Decimal, localcontext

from dateutil.relativedelta import relativedelta
from sql import Window
from sql.functions import NthValue

from trytond.i18n import gettext
from trytond.model import (
    Check, DeactivableMixin, DigitsMixin, Index, ModelSQL, ModelView,
    SymbolMixin, Unique, fields)
from trytond.pool import Pool
from trytond.pyson import Eval, If
from trytond.rpc import RPC
from trytond.transaction import Transaction

from .ecb import RatesNotAvailableError, get_rates
from .exceptions import RateError
from .ir import rate_decimal

logger = logging.getLogger(__name__)


ROUNDING_OPPOSITES = {
    decimal.ROUND_HALF_EVEN: decimal.ROUND_HALF_EVEN,
    decimal.ROUND_HALF_UP: decimal.ROUND_HALF_DOWN,
    decimal.ROUND_HALF_DOWN: decimal.ROUND_HALF_UP,
    decimal.ROUND_UP: decimal.ROUND_DOWN,
    decimal.ROUND_DOWN: decimal.ROUND_UP,
    decimal.ROUND_CEILING: decimal.ROUND_FLOOR,
    decimal.ROUND_FLOOR: decimal.ROUND_CEILING,
    }


class Currency(
        SymbolMixin, DigitsMixin, DeactivableMixin, ModelSQL, ModelView):
    'Currency'
    __name__ = 'currency.currency'
    name = fields.Char('Name', required=True, translate=True,
        help="The main identifier of the currency.")
    symbol = fields.Char(
        "Symbol", size=10, strip=False,
        help="The symbol used for currency formating.")
    code = fields.Char('Code', size=3, required=True,
        help="The 3 chars ISO currency code.")
    numeric_code = fields.Char('Numeric Code', size=3,
        help="The 3 digits ISO currency code.")
    rate = fields.Function(fields.Numeric(
            "Current rate", digits=(rate_decimal * 2, rate_decimal)),
        'get_rate')
    rates = fields.One2Many('currency.currency.rate', 'currency', 'Rates',
        help="Add floating exchange rates for the currency.")
    rounding = fields.Numeric('Rounding factor', required=True,
        digits=(12, Eval('digits', 6)),
        help="The minimum amount which can be represented in this currency.")
    digits = fields.Integer("Digits", required=True,
        help="The number of digits to display after the decimal separator.")

    @classmethod
    def __setup__(cls):
        super(Currency, cls).__setup__()
        cls._order.insert(0, ('code', 'ASC'))
        cls.__rpc__.update({
                'compute': RPC(instantiate=slice(0, 3, 2)),
                })

    @classmethod
    def __register__(cls, module_name):
        pool = Pool()
        Data = pool.get('ir.model.data')
        data = Data.__table__()
        cursor = Transaction().connection.cursor()

        super(Currency, cls).__register__(module_name)

        table_h = cls.__table_handler__(module_name)

        # Migration from 5.2: remove country data
        cursor.execute(*data.delete(where=(data.module == 'currency')
                & (data.model == cls.__name__)))

        # Migration from 6.6: remove required on symbol
        table_h.not_null_action('symbol', 'remove')

    @staticmethod
    def default_rounding():
        return Decimal('0.01')

    @staticmethod
    def default_digits():
        return 2

    @classmethod
    def check_xml_record(cls, records, values):
        pass

    @classmethod
    def search_global(cls, text):
        for record, rec_name, icon in super(Currency, cls).search_global(text):
            icon = icon or 'tryton-currency'
            yield record, rec_name, icon

    @classmethod
    def search_rec_name(cls, name, clause):
        currencies = None
        field = None
        for field in ('code', 'numeric_code'):
            currencies = cls.search([(field,) + tuple(clause[1:])], limit=1)
            if currencies:
                break
        if currencies:
            return [(field,) + tuple(clause[1:])]
        return [(cls._rec_name,) + tuple(clause[1:])]

    @fields.depends('rates')
    def on_change_with_rate(self):
        now = datetime.date.today()
        closer = datetime.date.min
        res = Decimal(0)
        for rate in self.rates or []:
            date = getattr(rate, 'date', None) or now
            if date <= now and date > closer:
                res = rate.rate
                closer = date
        return res

    @staticmethod
    def get_rate(currencies, name):
        '''
        Return the rate at the date from the context or the current date
        '''
        Rate = Pool().get('currency.currency.rate')
        Date = Pool().get('ir.date')

        res = {}
        date = Transaction().context.get('date', Date.today())
        for currency in currencies:
            rates = Rate.search([
                    ('currency', '=', currency.id),
                    ('date', '<=', date),
                    ], limit=1, order=[('date', 'DESC')])
            if rates:
                res[currency.id] = rates[0].id
            else:
                res[currency.id] = 0
        rate_ids = [x for x in res.values() if x]
        rates = Rate.browse(rate_ids)
        id2rate = {}
        for rate in rates:
            id2rate[rate.id] = rate
        for currency_id in res.keys():
            if res[currency_id]:
                res[currency_id] = id2rate[res[currency_id]].rate
        return res

    def round(self, amount, rounding=decimal.ROUND_HALF_EVEN, opposite=False):
        'Round the amount depending of the currency'
        if opposite:
            rounding = ROUNDING_OPPOSITES[rounding]
        return self._round(amount, self.rounding, rounding)

    @classmethod
    def _round(cls, amount, factor, rounding):
        if not factor:
            return amount
        with localcontext() as ctx:
            ctx.prec = max(ctx.prec, (amount / factor).adjusted() + 1)
            # Divide and multiple by factor for case factor is not 10En
            result = (amount / factor).quantize(Decimal('1.'),
                    rounding=rounding) * factor
        return Decimal(result)

    def is_zero(self, amount):
        'Return True if the amount can be considered as zero for the currency'
        if not self.rounding:
            return not amount
        return abs(self.round(amount)) < abs(self.rounding)

    @classmethod
    def compute(cls, from_currency, amount, to_currency, round=True):
        '''
        Take a currency and an amount
        Return the amount to the new currency
        Use the rate of the date of the context or the current date
        '''
        Date = Pool().get('ir.date')
        Lang = Pool().get('ir.lang')
        from_currency = cls(int(from_currency))
        to_currency = cls(int(to_currency))

        if to_currency == from_currency:
            if round:
                return to_currency.round(amount)
            else:
                return amount
        if (not from_currency.rate) or (not to_currency.rate):
            date = Transaction().context.get('date', Date.today())
            if not from_currency.rate:
                name = from_currency.name
            else:
                name = to_currency.name

            lang = Lang.get()
            raise RateError(gettext('currency.msg_no_rate',
                    currency=name,
                    date=lang.strftime(date)))
        if round:
            return to_currency.round(
                amount * to_currency.rate / from_currency.rate)
        else:
            return amount * to_currency.rate / from_currency.rate

    @classmethod
    def currency_rate_sql(cls):
        "Return a SQL query with currency, rate, start_date and end_date"
        pool = Pool()
        Rate = pool.get('currency.currency.rate')
        transaction = Transaction()
        database = transaction.database

        rate = Rate.__table__()
        if database.has_window_functions():
            window = Window(
                [rate.currency],
                order_by=[rate.date.asc],
                frame='ROWS', start=0, end=1)
            # Use NthValue instead of LastValue to get NULL for the last row
            end_date = NthValue(rate.date, 2, window=window)
        else:
            next_rate = Rate.__table__()
            end_date = next_rate.select(
                next_rate.date,
                where=(next_rate.currency == rate.currency)
                & (next_rate.date > rate.date),
                order_by=[next_rate.date.asc],
                limit=1)

        query = (rate
            .select(
                rate.currency.as_('currency'),
                rate.rate.as_('rate'),
                rate.date.as_('start_date'),
                end_date.as_('end_date'),
                ))
        return query

    def get_symbol(self, sign, symbol=None):
        pool = Pool()
        Lang = pool.get('ir.lang')
        lang = Lang.get()
        symbol, position = super().get_symbol(sign, symbol=symbol)
        if not symbol:
            symbol = self.code
        if ((sign < 0 and lang.n_cs_precedes)
                or (sign >= 0 and lang.p_cs_precedes)):
            position = 0
        return symbol, position


class CurrencyRate(ModelSQL, ModelView):
    "Currency Rate"
    __name__ = 'currency.currency.rate'
    date = fields.Date(
        "Date", required=True,
        help="From when the rate applies.")
    rate = fields.Numeric(
        "Rate", digits=(rate_decimal * 2, rate_decimal), required=1,
        help="The floating exchange rate used to convert the currency.")
    currency = fields.Many2One('currency.currency', 'Currency',
            ondelete='CASCADE',
        help="The currency on which the rate applies.")

    @classmethod
    def __setup__(cls):
        super().__setup__()
        cls.__access__.add('currency')
        t = cls.__table__()
        cls._sql_constraints = [
            ('date_currency_uniq', Unique(t, t.date, t.currency),
                'currency.msg_currency_unique_rate_date'),
            ('check_currency_rate', Check(t, t.rate >= 0),
                'currency.msg_currency_rate_positive'),
            ]
        cls._sql_indexes.add(
            Index(
                t,
                (t.currency, Index.Equality()),
                (t.date, Index.Range()),
                order='DESC'))
        cls._order.insert(0, ('date', 'DESC'))

    @staticmethod
    def default_date():
        Date = Pool().get('ir.date')
        return Date.today()

    @classmethod
    def check_xml_record(cls, records, values):
        pass

    def get_rec_name(self, name):
        return str(self.date)


class CronFetchError(Exception):
    pass


class Cron(ModelSQL, ModelView):
    "Currency Cron"
    __name__ = 'currency.cron'

    source = fields.Selection(
        [('ecb', "European Central Bank")],
        "Source", required=True,
        help="The external source for rates.")
    frequency = fields.Selection([
            ('daily', "Daily"),
            ('weekly', "Weekly"),
            ('monthly', "Monthly"),
            ], "Frequency", required=True,
        help="How frequently rates must be updated.")
    weekday = fields.Many2One(
        'ir.calendar.day', "Day of Week",
        states={
            'required': Eval('frequency') == 'weekly',
            'invisible': Eval('frequency') != 'weekly',
            })
    day = fields.Integer(
        "Day of Month",
        domain=[If(Eval('frequency') == 'monthly',
                [('day', '>=', 1), ('day', '<=', 31)],
                [('day', '=', None)]),
            ],
        states={
            'required': Eval('frequency') == 'monthly',
            'invisible': Eval('frequency') != 'monthly',
            })
    currency = fields.Many2One(
        'currency.currency', "Currency", required=True,
        help="The base currency to fetch rate.")
    currencies = fields.Many2Many(
        'currency.cron-currency.currency', 'cron', 'currency', "Currencies",
        help="The currencies to update the rate.")
    last_update = fields.Date("Last Update", required=True)

    @classmethod
    def __setup__(cls):
        super().__setup__()
        cls._buttons.update({
                'run': {},
                })

    @classmethod
    def default_frequency(cls):
        return 'monthly'

    @classmethod
    def default_day(cls):
        return 1

    @classmethod
    def default_last_update(cls):
        pool = Pool()
        Date = pool.get('ir.date')
        return Date.today()

    @classmethod
    @ModelView.button
    def run(cls, crons):
        cls.update(crons)

    @classmethod
    def update(cls, crons=None):
        pool = Pool()
        Rate = pool.get('currency.currency.rate')
        if crons is None:
            crons = cls.search([])
        rates = []
        for cron in crons:
            rates.extend(cron._update())
        Rate.save(rates)
        cls.save(crons)

    def _update(self):
        limit = self.limit_update()
        date = self.next_update()
        while date <= limit:
            try:
                yield from self._rates(date)
            except CronFetchError:
                logger.warning("Could not fetch rates temporary")
                if date >= datetime.date.today():
                    break
            except Exception:
                logger.error("Fail to fetch rates", exc_info=True)
                break
            self.last_update = date
            date = self.next_update()

    def next_update(self):
        return self.last_update + self.delta()

    def limit_update(self):
        pool = Pool()
        Date = pool.get('ir.date')
        return Date.today()

    def delta(self):
        if self.frequency == 'daily':
            delta = relativedelta(days=1)
        elif self.frequency == 'weekly':
            delta = relativedelta(weeks=1, weekday=int(self.weekday.index))
        elif self.frequency == 'monthly':
            delta = relativedelta(months=1, day=self.day)
        else:
            delta = relativedelta()
        return delta

    def _rates(self, date, rounding=None):
        pool = Pool()
        Rate = pool.get('currency.currency.rate')
        values = getattr(self, 'fetch_%s' % self.source)(date)

        exp = Decimal(Decimal(1) / 10 ** Rate.rate.digits[1])
        rates = Rate.search([
                ('date', '=', date),
                ])
        code2rates = {r.currency.code: r for r in rates}

        def get_rate(currency):
            if currency.code in code2rates:
                rate = code2rates[currency.code]
            else:
                rate = Rate(date=date, currency=currency)
            return rate

        rate = get_rate(self.currency)
        rate.rate = Decimal(1).quantize(exp, rounding=rounding)
        yield rate

        for currency in self.currencies:
            if currency.code not in values:
                continue
            value = values[currency.code]
            if not isinstance(value, Decimal):
                value = Decimal(value)
            rate = get_rate(currency)
            rate.rate = value.quantize(exp, rounding=rounding)
            yield rate

    def fetch_ecb(self, date):
        try:
            return get_rates(self.currency.code, date)
        except RatesNotAvailableError as e:
            raise CronFetchError() from e


class Cron_Currency(ModelSQL):
    "Currency Cron - Currency"
    __name__ = 'currency.cron-currency.currency'

    cron = fields.Many2One(
        'currency.cron', "Cron",
        required=True, ondelete='CASCADE')
    currency = fields.Many2One(
        'currency.currency', "Currency",
        required=True, ondelete='CASCADE')



#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK
# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import gettext
import os
import sys
from argparse import ArgumentParser

import pycountry

try:
    import argcomplete
except ImportError:
    argcomplete = None

try:
    from tqdm import tqdm
except ImportError:
    tqdm = None

try:
    from proteus import Model, config
except ImportError:
    prog = os.path.basename(sys.argv[0])
    sys.exit("proteus must be installed to use %s" % prog)


def _progress(iterable, **kwargs):
    if tqdm:
        return tqdm(iterable, disable=None, **kwargs)
    else:
        return iterable


def _get_language_codes():
    Language = Model.get('ir.lang')
    languages = Language.find([('translatable', '=', True)])
    for l in languages:
        yield l.code


def _remove_forbidden_chars(name):
    from trytond.tools import remove_forbidden_chars
    return remove_forbidden_chars(name)


def get_currencies():
    Currency = Model.get('currency.currency')
    return {c.code: c for c in Currency.find([])}


def update_currencies(currencies):
    print("Update currencies", file=sys.stderr)
    Currency = Model.get('currency.currency')

    records = []
    for currency in _progress(pycountry.currencies):
        code = currency.alpha_3
        if code in currencies:
            record = currencies[code]
        else:
            record = Currency(code=code)
        record.name = _remove_forbidden_chars(currency.name)
        record.numeric_code = currency.numeric
        records.append(record)

    Currency.save(records)
    return {c.code: c for c in records}


def translate_currencies(currencies):
    Currency = Model.get('currency.currency')

    current_config = config.get_config()
    for code in _get_language_codes():
        try:
            gnutranslation = gettext.translation(
                'iso4217', pycountry.LOCALES_DIR, languages=[code])
        except IOError:
            continue
        print("Update currencies %s" % code, file=sys.stderr)
        with current_config.set_context(language=code):
            records = []
            for currency in _progress(pycountry.currencies):
                record = Currency(currencies[currency.alpha_3].id)
                record.name = _remove_forbidden_chars(
                    gnutranslation.gettext(currency.name))
                records.append(record)
            Currency.save(records)


def main(database, config_file=None):
    config.set_trytond(database, config_file=config_file)
    with config.get_config().set_context(active_test=False):
        do_import()


def do_import():
    currencies = get_currencies()
    currencies = update_currencies(currencies)
    translate_currencies(currencies)


def run():
    parser = ArgumentParser()
    parser.add_argument('-d', '--database', dest='database', required=True)
    parser.add_argument('-c', '--config', dest='config_file',
        help='the trytond config file')
    if argcomplete:
        argcomplete.autocomplete(parser)

    args = parser.parse_args()
    main(args.database, args.config_file)


if __name__ == '__main__':
    run()



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

import datetime as dt
import ssl
import sys

try:
    from lxml import etree as ET
except ImportError:
    try:
        import xml.etree.cElementTree as ET
    except ImportError:
        import xml.etree.ElementTree as ET

from decimal import Decimal
from urllib.error import HTTPError
from urllib.request import urlopen

_URL = 'https://www.ecb.europa.eu/stats/eurofxref/'
_URL_TODAY = _URL + 'eurofxref-daily.xml'
_URL_90 = _URL + 'eurofxref-hist-90d.xml'
_URL_HIST = _URL + 'eurofxref-hist.xml'
_START_DATE = dt.date(1999, 1, 4)
_CUBE_TAG = '{http://www.ecb.int/vocabulary/2002-08-01/eurofxref}Cube'


class RatesNotAvailableError(Exception):
    pass


class UnsupportedCurrencyError(Exception):
    pass


def _parse_time(time):
    return dt.datetime.strptime(time, '%Y-%m-%d').date()


def _find_time(source, time=None):
    for _, element in ET.iterparse(source):
        if element.tag == _CUBE_TAG and 'time' in element.attrib:
            if time and _parse_time(element.attrib.get('time')) <= time:
                return element
            elif time is None:
                return element
            element.clear()


def get_rates(currency='EUR', date=None):
    if date is None:
        date = dt.date.today()
    if date < _START_DATE:
        date = _START_DATE
    context = ssl.create_default_context()
    try:
        with urlopen(_URL_TODAY, context=context) as response:
            element = _find_time(response)
            last_date = _parse_time(element.attrib['time'])
            if last_date < date:
                raise RatesNotAvailableError()
            elif last_date == date:
                return _compute_rates(element, currency, date)

        if last_date - date < dt.timedelta(days=90):
            url = _URL_90
        else:
            url = _URL_HIST
        with urlopen(url, context=context) as response:
            element = _find_time(response, date)
            if element is None and url == _URL_90:
                with urlopen(_URL_HIST, context=context) as response:
                    element = _find_time(response, date)
            return _compute_rates(element, currency, date)
    except HTTPError as e:
        raise RatesNotAvailableError() from e


def _compute_rates(element, currency, date):
    currencies = {}
    for cur in element:
        currencies[cur.attrib['currency']] = Decimal(cur.attrib['rate'])
    if currency != 'EUR':
        currencies['EUR'] = Decimal(1)
        try:
            base_rate = currencies.pop(currency)
        except KeyError:
            raise UnsupportedCurrencyError(f'{currency} is not available')
        for cur, rate in currencies.items():
            currencies[cur] = (rate / base_rate).quantize(Decimal('.0001'))
    return currencies


if __name__ == '__main__':
    currency = 'EUR'
    if len(sys.argv) > 1:
        currency = sys.argv[1]
    date = None
    if len(sys.argv) > 2:
        date = dt.datetime.strptime(sys.argv[2], '%Y-%m-%d').date()
    print(get_rates(currency, date))
